/*
 * tokyocabinet.c
 */

#include "tokyocabinet.h"

#define TCCALL(call) ((call) ? SCM_TRUE : SCM_FALSE)
// #define DEBUG_GAUCHE_TOKYOCABINET

/* These variables are for temporary use. */
static TCXSTR * xkey;
static TCXSTR * xvalue;


static void tcbdb_finalize(ScmObj obj, void *data)
{
    ScmRawTcbdb *bdb = SCM_RAW_TCBDB(obj);
    tcbdbclose(bdb->bdb);
}
static void tchdb_finalize(ScmObj obj, void *data)
{
    ScmRawTchdb *hdb = SCM_RAW_TCHDB(obj);
    tchdbclose(hdb->hdb);
}



static void tcbdb_print(ScmObj obj, ScmPort *out, ScmWriteContext *ctx){
    Scm_Printf(out, "#<tcbdb-file>");
}
static void tchdb_print(ScmObj obj, ScmPort *out, ScmWriteContext *ctx){
    Scm_Printf(out, "#<tchdb-file>");
}

ScmObj Scm_tcbdb_ecode(ScmRawTcbdb* bdb){
    return SCM_MAKE_INT(tcbdbecode(SCM_RAW_TCBDB(bdb)->bdb));
}
ScmObj Scm_tchdb_ecode(ScmRawTchdb* hdb){
    return SCM_MAKE_INT(tchdbecode(SCM_RAW_TCHDB(hdb)->hdb));
}



ScmObj Scm_tcbdb_new(void){
    ScmRawTcbdb *ret = SCM_NEW(ScmRawTcbdb);
    SCM_SET_CLASS(ret, SCM_CLASS_RAW_TCBDB);
    Scm_RegisterFinalizer(SCM_OBJ(ret), tcbdb_finalize, NULL);
    ret->bdb = tcbdbnew();
    return SCM_OBJ(ret);
}
ScmObj Scm_tchdb_new(void){
    ScmRawTchdb *ret = SCM_NEW(ScmRawTchdb);
    SCM_SET_CLASS(ret, SCM_CLASS_RAW_TCHDB);
    Scm_RegisterFinalizer(SCM_OBJ(ret), tchdb_finalize, NULL);
    ret->hdb = tchdbnew();
    return SCM_OBJ(ret);
}



ScmObj Scm_tcbdb_open(ScmRawTcbdb* bdb, ScmString* path, int omode){
    ScmObj ret = TCCALL(tcbdbopen(bdb->bdb
                                  , Scm_GetString(path)
                                  , omode));
}
ScmObj Scm_tchdb_open(ScmRawTchdb* hdb, ScmString* path, int omode){
    ScmObj ret = TCCALL(tchdbopen(hdb->hdb
                                  , Scm_GetString(path)
                                  , omode));
}



ScmObj Scm_tcbdb_close(ScmRawTcbdb* bdb){
    return TCCALL(tcbdbclose(bdb->bdb));
}
ScmObj Scm_tchdb_close(ScmRawTchdb* hdb){
    return TCCALL(tchdbclose(hdb->hdb));
}



ScmObj Scm_tchdb_put2(ScmRawTchdb* hdb, ScmString* key, ScmString* val){
    return TCCALL(tchdbput2(hdb->hdb
                            , Scm_GetString(key)
                            , Scm_GetString(val)));
}



ScmObj Scm_tchdb_get2(ScmRawTchdb* hdb, ScmString* key){
    char * val = tchdbget2(hdb->hdb
                           , Scm_GetString(key));
    if(val == NULL){
        return SCM_FALSE;
    }

    ScmObj ret = SCM_MAKE_STR(val);
    return ret;
}



ScmObj Scm_tcbdb_out2(ScmRawTcbdb* bdb, ScmString* key){
    return TCCALL(tcbdbout2(bdb->bdb
                            , Scm_GetString(key)));
}
ScmObj Scm_tchdb_out2(ScmRawTchdb* hdb, ScmString* key){
    return TCCALL(tchdbout2(hdb->hdb
                            , Scm_GetString(key)));
}


/*
  B+ tree specific API from here
 */
static void tcbdbcur_finalize(ScmObj obj, void *data)
{
    ScmTcbdbcur *cur = SCM_TCBDBCUR(obj);
    tcbdbcurdel(cur->cur);
}

static void tcbdbcur_print(ScmObj obj, ScmPort *out, ScmWriteContext *ctx){
    Scm_Printf(out, "#<tcbdb-cursor>");
}

ScmObj Scm_tcbdbcur_new(ScmRawTcbdb* bdb){
    ScmTcbdbcur *ret = SCM_NEW(ScmTcbdbcur);
    SCM_SET_CLASS(ret, SCM_CLASS_TCBDBCUR);
    Scm_RegisterFinalizer(SCM_OBJ(ret), tcbdbcur_finalize, NULL);
    ret->cur = tcbdbcurnew(bdb->bdb);
    return SCM_OBJ(ret);
}

ScmObj Scm_tcbdbcur_first(ScmTcbdbcur* cur){
    return TCCALL(tcbdbcurfirst(cur->cur));
}

ScmObj Scm_tcbdbcur_last(ScmTcbdbcur* cur){
    return TCCALL(tcbdbcurlast(cur->cur));
}

ScmObj Scm_tcbdbcur_prev(ScmTcbdbcur* cur){
    return TCCALL(tcbdbcurprev(cur->cur));
}

ScmObj Scm_tcbdbcur_next(ScmTcbdbcur* cur){
    return TCCALL(tcbdbcurnext(cur->cur));
}

ScmObj Scm_tcbdbcur_key2(ScmTcbdbcur* cur){
    char * str = tcbdbcurkey2(cur->cur);
    if(str == NULL){
        return SCM_FALSE;
    }
    ScmObj ret = SCM_MAKE_STR(str);
    return ret;
}

ScmObj Scm_tcbdbcur_val2(ScmTcbdbcur* cur){
    char * str = tcbdbcurval2(cur->cur);
    if(str == NULL){
        return SCM_FALSE;
    }
    ScmObj ret = SCM_MAKE_STR(str);
    return ret;
}

/*
  B+ tree specific API until here
 */


/*
  Hash specific API from here
 */



/*
  Hash specific API until here
 */

extern ScmObj Scm_tchdb_iterinit(ScmRawTchdb* hdb){
#ifdef DEBUG_GAUCHE_TOKYOCABINET
    puts("Scm_tchdb_iterinit");
#endif
    return TCCALL(tchdbiterinit(hdb->hdb));
}

extern ScmObj Scm_tchdb_iternext2(ScmRawTchdb* hdb){
    char * key = tchdbiternext2(hdb->hdb);
    if (key == NULL){
        return SCM_FALSE;
    }

    /*
     *key is newly malloced in tchdbiternext2, so
     SCM_MAKE_STR is appropriate, because it does not
     copy its argument (STM_MAKE_STR_COPYING does)
     */
    return SCM_MAKE_STR(key);
}



extern ScmObj Scm_tchdb_iternext3(ScmRawTchdb* hdb){
#ifdef DEBUG_GAUCHE_TOKYOCABINET
    printf("Scm_tchdb_iternext3( 0x%x )", (uintptr_t) hdb);
#endif 
    ScmObj ret_iternext, key, val;

#ifdef DEBUG_GAUCHE_TOKYOCABINET
    printf("call tchdbiternext3( 0x%x , 0x%x , 0x%x )"
           , (uintptr_t) hdb->hdb
           , (uintptr_t) xkey
           , (uintptr_t) xvalue);
#endif
    ret_iternext = TCCALL(tchdbiternext3(hdb->hdb, xkey, xvalue));
    key = SCM_MAKE_STR_COPYING(tcxstrptr(xkey));
    val = SCM_MAKE_STR_COPYING(tcxstrptr(xvalue));
    return Scm_Values3(ret_iternext, key, val);
}



/*
 * Module initialization function.
 */
extern void Scm_Init_tokyocabinetlib(ScmModule*);

void Scm_Init_tokyocabinet(void)
{
    ScmModule *mod;

    xkey = tcxstrnew();
    xvalue = tcxstrnew();

    /* Register this DSO to Gauche */
    SCM_INIT_EXTENSION(tokyocabinet);

    /* Create the module if it doesn't exist yet. */
    mod = SCM_MODULE(SCM_FIND_MODULE("dbm.tokyocabinet", TRUE));
    Scm_InitStaticClass(&Scm_RawTcbdbClass, "<raw-tcbdb>", mod, NULL, 0);
    Scm_InitStaticClass(&Scm_RawTchdbClass, "<raw-tchdb>", mod, NULL, 0);

    /* Register stub-generated procedures */
    Scm_Init_tokyocabinetlib(mod);
}
